/***
 *  Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#ifndef WORK_LAYERPREVIEW_H
#define WORK_LAYERPREVIEW_H

#include "commtypes.h"
#include "commdefs.h"
#include "refcounter.h"

//#ifdef QT_OPENGL_LIB
#   include "../../glew/include/GL/glew.h"

#include "iArray.h"
#include "iCanvasLayer.h"
#include "iOpenGLHelper.h"
#include "iwork.h"
#include "worker_params.h"

#include <QObject>

#define PREVIEW_WIDTH       320
#define PREVIEW_HEIGHT      240


class Work_LayerPreview : public QObject,
                          public IWork,
                          public CRefCounter {
    Q_OBJECT
    _VERBOSEOBJECT

signals:
    /**
     * Broadcast our result through the message.
     */
    void previewUpdated (QImage * image);

public:
    Work_LayerPreview (ICanvas       * canvas,
                       ICanvasLayer  * layer,
                       IOpenGLHelper * helper,
                       int  priority,
                       bool persistent,
                       const sWorkerParameters * params);

    /**
     * Perform the work.
     */
    virtual bool execute (void);

    /**
     * Get the task priority.
     */
    virtual int priority (void) const {
        return m_Priority;
    }

    /**
     * Get the task persistence flag.
     */
    virtual bool isPersisitent (void) const {
        return m_Persistent;
    }

    /**
     * Check if the task is active now.
     */
    virtual bool isActive (void) const {
        return m_Active;
    }

    /**
     * Activate or deactivate the task.
     */
    virtual void setActive (bool state) {
        m_Active = state;
    }

    /**
     * Get the task execution interval in ms.
     */
    virtual long period (void) const {
        return 5000L;
    }

    /**
     * Pass the connector interface
     * to the object.
     */
    virtual void connect (IConnector * caller);

    /**
     * A callback, called by connector when
     * the requested data is recieved.
     * @param param Caller-define parameter associated with request.
     * @param length The length of the data string in bytes.
     * @param data Data returned. This buffer will be invalid after the function exit.
     */
    virtual void complitedRequest (int param, int length, void * data);

    /******** IVERBOSEOBJECT IMPLANTS ************/

    /**
     * Returns the description of last error
     * or NULL if no error occured.
     */
    virtual const char * lastError (void) const {
        return m_LastError;
    }

    /**
     * Attach the logger to an object.
     */
    virtual void attachLogger (ILogger * logger) {
        SAFE_RELEASE (m_Logger);
        if (logger) {
            m_Logger = logger;
            m_Logger->addRef ();
        }
    }

    /**
     * Enable of disable logging.
     */
    virtual void enableLogging (bool enable) {
        m_LoggingEnabled = enable;
    }

    /**
     * Check if the logging is enabled.
     */
    virtual bool loggingEnabled (void) {
        return m_LoggingEnabled;
    }

    /******** ISPAWNEDOBJECT IMPLANTS ************/

    /**
     * Get the classname for this object.
     */
    virtual const char * myClassName (void) {
        return "Client::BGTask";
    }

    /**
     * Get the Id of current allocated object.
     */
    virtual int objectId (void) {
        return m_Id;
    }

    /**
     * Increase the object's
     * reference counter.
     */
    virtual void addRef (void) {
        CRefCounter::addRef ();
    }

    /**
     * Release the object.
     */
    virtual void release (void) {
        CRefCounter::release ();
    }

    /**
     * Ask the object to emit all
     * notification signals if any.
     */
    virtual void pokeObject (void) {
    }

protected:
    int               m_Id;
    int               m_Priority;         /**< The priority of this task. */
    volatile bool     m_Active;           /**< Activity flag.             */
    bool              m_Persistent;       /**< Persistence flag.          */
    ICanvasLayer    * m_Layer;            /**< Target layer.              */
    ICanvas         * m_Canvas;           /**< Canvas object.             */
    IOpenGLHelper   * m_Helper;
    IConnector      * m_Connector;
    IArray          * m_Preview;          /**< Render target.              */

    QImage          * m_Previewi;

    /**
     * Shared parameters block.
     */
    const sWorkerParameters * m_Params;

    virtual ~Work_LayerPreview (void);

    virtual void suicide (void) {
        delete this;
    }
};

#endif // WORK_LAYERPREVIEW_H
