/***************************************************************************
 *   Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>              *
 *                                                                         *
 *   This file is a part of QAquarelle project.                            *
 *                                                                         *
 *   QAquarelle is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef INSTRUMENT_BRUSHES_H
#define INSTRUMENT_BRUSHES_H

#include "commtypes.h"
#include "commdefs.h"
#include "refcounter.h"

#include "../../../glew/include/GL/glew.h"

#include <QtPlugin>
#include <QObject>

#define ALLOW_INTERFACE_DECLARATION

#include "instrument_brushes_global.h"
#include "iInstrument.h"
#include "iOpenGLHelper.h"
#include "iLogger.h"


class INSTRUMENT_BRUSHESSHARED_EXPORT Instrument_brushes : public QObject,
                                                           public IInstrumentFactory,
                                                           public CRefCounter {

    Q_OBJECT
    Q_INTERFACES (IInstrumentFactory)

    int m_Id;

    virtual void suicide (void) {
        // Static object?
        //delete this;
    }

protected:

    /**
     * External logger object.
     */
    ILogger * m_Logger;

public:
     Instrument_brushes (void);
    ~Instrument_brushes (void);

    /******* IINSTRUMENTFACTORY IMPLANTS ********/

    /**
     * Get the number of instruments
     * in this library.
     */
    virtual int instrumentsInLibrary (void) {
        return INSTRUMENTS_IN_PACK + INSTRUMENTS_IN_TOOLPACK;
    }

    /**
     * Get the classname of nth instrument
     * in library.
     */
    virtual const char * instrumentClassname (int index) {
        if (index >= 0 && index < instrumentsInLibrary ()) {
            extern const char * g_CLASSES_br[INSTRUMENTS_IN_PACK +
                                             INSTRUMENTS_IN_TOOLPACK];
            return g_CLASSES_br[index];
        } else
            return NULL;
    }

    /**
     * Attach the logger object that will
     * be used to log the messages for
     * all allocated instruments.
     */
    virtual void attachLogger (ILogger * logger) {
        SAFE_RELEASE (m_Logger);

        if (logger) {
            m_Logger = logger;
            m_Logger->addRef ();
        }
    }

    /**
     * Create an instrument by index.
     * @param index Instrument index, bounded by the value returned by instrumentsInLibrary ().
     * @param helper OpenGL helper object.
     */
    virtual IInstrument * create (int index, IOpenGLHelper * helper);

    /**
     * Create an instrument by classname.
     * @param index Instrument index, bounded by the value returned by instrumentsInLibrary ().
     * @param helper OpenGL helper object.
     */
    virtual IInstrument * create (const char * name, IOpenGLHelper * helper);

    /********** IPLUGIN IMPLANTS ************/

    /**
     * Get the name of this plugin.
     */
    virtual const char * name (void) {
        return "Brushes factory";
    }

    /**
     * Get the description of this plugin.
     */
    virtual const char * description (void) {
        return "A factory object for a Brushes instrument pack.";
    }

    /**
     * Get the author of this plugin.
     */
    virtual const char * author (void) {
        return "shadow-of-kyle";
    }

    /**
     * Get the number of custom parameters
     * accepted by the plugin.
     */
    virtual int customParametersCount (void) const {
        return 0;
    }

    /**
     * Get the name of custom parameter
     * with the given index.
     * /return Pointer to string of NULL if the index is out of range.
     */
    virtual const char * customParameterName (int index) const {
        Q_UNUSED (index);

        return NULL;
    }

    /**
     * Get the type of custom parameter
     * with the given index.
     */
    virtual ePLUGIN_PARAM_TYPE customParameterType (int index) const {
        Q_UNUSED (index);

        return PARAM_UNDEFINED;
    }

    /**
     * Get the type of custom parameter
     * with the given name.
     */
    virtual ePLUGIN_PARAM_TYPE customParameterType (const char * name) const {
        Q_UNUSED (name);

        return PARAM_UNDEFINED;
    }

    /**
     * Get the value of custom parameter with the given name.
     * /return TRUE if the value has been successfully retrieved.
     */
    virtual bool customParameterGet (const char * name, void * buffer) {
        Q_UNUSED (name);
        Q_UNUSED (buffer);

        return FALSE;
    }

    /**
     * Set the value of custom parameter with the given name.
     * /return TRUE if the value has been successfully set.
     */
    virtual bool customParameterSet (const char * name, const void * buffer) {
        Q_UNUSED (name);
        Q_UNUSED (buffer);

        return FALSE;
    }
    /**
     * Deinit all OpenGL-related data
     * structures to be prepared for a
     * context switching.
     */
    virtual void deInit (void) {
    }

    /******* ISPAWNEDOBJECT IMPLANTS *******/

    /**
     * Get the classname for this object.
     */
    virtual const char * myClassName (void) {
        return "Server::BrushesFactory";
    }

    /**
     * Get the Id of current allocated object.
     */
    virtual int objectId (void) {
        return m_Id;
    }

    /**
     * Increase the object's
     * reference counter.
     */
    virtual void addRef (void) {
        CRefCounter::addRef ();
    }

    /**
     * Release the object.
     */
    virtual void release (void) {
        CRefCounter::release ();
    }

    /**
     * Ask the object to emit all
     * notification signals if any.
     */
    virtual void pokeObject (void) {
    }
};

#endif // INSTRUMENT_BRUSHES_H
