/***
 *  Copyright (C) 2009 Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "simplebrush.h"
#include <math.h>
#include <stdlib.h>

#include <QImage>

#ifndef BUFFER_OFFSET
#   define BUFFER_OFFSET(x)     ((char*)0 + x)
#endif


const char * g_NAMES_br[INSTRUMENTS_IN_PACK +
                        INSTRUMENTS_IN_TOOLPACK] = {
    "Round small brush",
    "Flat small brush",
    "Pencil",
    "Eraser",
    "Eyedropper",
    "Selection"
};

const char * g_CLASSES_br[INSTRUMENTS_IN_PACK +
                          INSTRUMENTS_IN_TOOLPACK] = {
    "Server::BR_Round_sm",
    "Server::BR_Flat_sm",
    "Server::Pencil",
    "Server::Eraser",
    "Server::Eyedropper",
    "Server::Selection"
};

static const float g_ZERO[4] = {
    0.0f, 0.0f, 0.0f, 0.0f
};
static const float g_GREY[4] = {
    0.3f, 0.3f, 0.3f, 1.0f
};

void BrushLock::release (void) {
    m_Locked = FALSE;
    if (m_Parent)
        ((BrushInstrument*)m_Parent)->slotUnlocked (this);
}

BrushInstrument::BrushInstrument (int type,
                                  IInstrumentFactory * factory,
                                  IOpenGLHelper      * helper,
                                  ILogger            * logger) {
    m_Type    = type;
    m_Logger  = logger;
    m_Helper  = helper;
    m_Factory = factory;
    m_LastX   = 0;
    m_LastY   = 0;

    m_Id = ISpawnedObject::allocateId ();
    ISpawnedObject::_register (this);

    m_Changed        = TRUE;
    m_Stroke         = FALSE;
    m_PositionCached = FALSE;

    if (m_Logger)  m_Logger->addRef ();
    if (m_Helper)  m_Helper->addRef ();
    if (m_Factory) m_Factory->addRef ();

    m_StrokeAngle      = 0.0f;
    m_Pressure         = 0.0f;
    m_WaterStorage     = 0.5f;
    m_UpdateRate_suck  = 0.3f;
    m_UpdateRate_bleed = 0.1f;
    m_Hardness         = 0.5f;

    // Reset the locking slots.
    for (int i = 0; i < SLOTS_NUMBER; i++) {
        m_LockSlots[i].m_Locked         = FALSE;
        m_LockSlots[i].m_Valid          = FALSE;
        m_LockSlots[i].m_PrimitiveColor = NULL;
        m_LockSlots[i].m_PrimitiveMask  = NULL;
        m_LockSlots[i].m_Sensitivity    = 0.0f;
        m_LockSlots[i].m_BleedingRate   = 1.0f;
        m_LockSlots[i].m_SuckingRate    = 0.0f;
        m_LockSlots[i].m_ScalingFactor  = 1.0f;
        m_LockSlots[i].m_Reservoir      = 0.0f;
        m_LockSlots[i].m_Channel        = NULL;
        m_LockSlots[i].m_Parent         = this;
        m_LockSlots[i].m_ChannelIndex   = i;

        m_LockSlots[i].m_ComponentMask[0] = +0.0f;
        m_LockSlots[i].m_ComponentMask[1] = +0.0f;
        m_LockSlots[i].m_ComponentMask[2] = +0.0f;
        m_LockSlots[i].m_ComponentMask[3] = +1.0f;

        if (m_Type == TYPE_PENCIL) {
            m_LockSlots[i].m_NoiseParams.frequency   = 160.0f;
            m_LockSlots[i].m_NoiseParams.amplitude   = 0.30f;
            m_LockSlots[i].m_NoiseParams.persistence = 1.0f;
            m_LockSlots[i].m_SuckingRate             = 0.00f;
            m_LockSlots[i].m_Affection               = DATAPLANE_WET;//DRY;
        } else {
            m_LockSlots[i].m_NoiseParams.frequency   = 1.00f;
            m_LockSlots[i].m_NoiseParams.amplitude   = 0.05f;
            m_LockSlots[i].m_NoiseParams.persistence = 0.00f;
            m_LockSlots[i].m_Affection               = DATAPLANE_WET;
            if (m_Type != TYPE_ERASER) {
                m_LockSlots[i].m_SuckingRate = 0.4f;
                m_LockSlots[i].m_Reservoir   = 0.7f;
                m_LockSlots[i].m_Affection  |= DATAPLANE_FLOW;
            } else {
                m_LockSlots[i].m_Affection |= DATAPLANE_DRY;
                m_LockSlots[i].m_Affection |= DATAPLANE_BG;
                m_LockSlots[i].m_Affection |= DATAPLANE_FLOW;

                // Invert the flow direction for eraser.
                m_LockSlots[i].m_ComponentMask[3] *= -1.0f;
            }
        }
    }

    buildTuftModel ();
    m_nVertices = 0;

    // OpenGL data.
    //
    m_GLinited = FALSE;
    m_VBOValid = VALID_NOTHING;

    m_Mask    = NULL;
    m_MaskCtl = NULL;
    m_Color   = NULL;

    // Load the instrument image.
    switch (m_Type) {
        case TYPE_BRUSH_ROUND:
            m_Image.load (":/res/round.png");
            break;
        case TYPE_BRUSH_FLAT:
            m_Image.load (":/res/flat.png");
            break;
        case TYPE_PENCIL:
            m_Image.load (":/res/pencil.png");
            break;
        case TYPE_ERASER:
            m_Image.load (":/res/eraser.png");
            break;
        default:
            break;
    }

    if (m_Type != TYPE_PENCIL) {
        memcpy (m_AbsorbingAvg,  g_ZERO, sizeof(float) * 4);
        memcpy (m_ScatteringAvg, g_ZERO, sizeof(float) * 4);
    } else {
        memcpy (m_AbsorbingAvg,  g_GREY, sizeof(float) * 4);
        memcpy (m_ScatteringAvg, g_ZERO, sizeof(float) * 4);
    }
}

void BrushInstrument::buildTuftModel (void) {
    // Build the spine.
    sTuftNode * n00 = m_Tuft.addNode (+0.0f, +2.0f, -0.0f, TRUE,  TRUE);
    sTuftNode * n01 = m_Tuft.addNode (+0.0f, +1.5f, -0.0f, TRUE,  TRUE);
    sTuftNode * n02 = m_Tuft.addNode (+0.0f, +1.0f, -0.0f, TRUE,  FALSE);
    sTuftNode * n03 = m_Tuft.addNode (+0.0f, +0.5f, -0.0f, TRUE,  FALSE);
    sTuftNode * n04 = m_Tuft.addNode (+0.0f, +0.2f, -0.0f, TRUE,  FALSE);
    sTuftNode * n05 = m_Tuft.addNode (+0.0f, +0.0f, -0.0f, TRUE,  FALSE);

    m_Tuft.linkNodes (n00, n01);
    m_Tuft.linkNodes (n01, n02);
    m_Tuft.linkNodes (n02, n03);
    m_Tuft.linkNodes (n03, n04);
    m_Tuft.linkNodes (n04, n05);

    // Tip.
    m_Tip = n05;

    if (m_Type != TYPE_BRUSH_ROUND) {
        m_Tuft.setSize       (0.80f);
        m_Tuft.setElastisity (0.02f);
    }
}

void BrushInstrument::initVertexBuffer (void) {
    glGenBuffersARB      (1, m_VBuf);

    // Bind and fill vertex data.
    glBindBufferARB (GL_ARRAY_BUFFER_ARB, m_VBuf[0]);
    glBufferDataARB (GL_ARRAY_BUFFER_ARB, MAX_PRIMITIVE_VERTICES * 5 * sizeof(float), NULL, GL_STREAM_DRAW_ARB);
    glBindBufferARB (GL_ARRAY_BUFFER_ARB, 0);

    glBindBufferARB (GL_ARRAY_BUFFER_ARB, 0);
}

void BrushInstrument::initBrushTexture_flat (IArray * mask, float randomness, bool alpha) {
    PROFILE_FUNC();
    //srand (m_LastX * m_LastY);  // FIXME Not needed anymore?

    if (mask) {
        float * p = (float*) mask->lock_pointer ();
        if (p) {
            const int w  = mask->width  ();
            const int h  = mask->height ();
            const int cx = w / 2;
            const int cy = h / 2;
            for (int y = 0; y < h; y++) {
                for (int x = 0; x < w; x++, p++) {
                    const float dx = ((float) abs (x - cx)) / cx;
                    const float dy = ((float) abs (y - cy)) / cy;
                    const float r  = 1.0f - (randomness * rand ()) / RAND_MAX;
                    if (r  > 0.5f &&
                        dx < 0.2f &&
                        dy < 0.8f) {
                        if (alpha)
                            *p = 1.0f - 0.5f * (1.0f - r);
                        else
                            *p = 1.0f;
                    } else
                        *p = 0.0f;
                }
            }

            mask->unlock ();
        }
    }
}

void BrushInstrument::initBrushTexture_round (IArray * mask, float randomness, bool alpha) {
    PROFILE_FUNC();
    //srand (m_LastX * m_LastY);  // FIXME Not needed anymore?

    if (mask) {
        float * p = (float*) mask->lock_pointer ();
        if (p) {
            const int w  = mask->width  ();
            const int h  = mask->height ();
            const int cx = w / 2;
            const int cy = h / 2;
            for (int y = 0; y < h; y++) {
                for (int x = 0; x < w; x++, p++) {
                    const float dx = ((float) abs (x - cx)) / cx;
                    const float dy = ((float) abs (y - cy)) / cy;
                    const float r  = 1.0f - (randomness * rand ()) / RAND_MAX;
                    const float R  = sqrtf (dx * dx + dy * dy);
                    if (r > 0.5f &&
                        R < 0.8f) {
                        if (alpha)
                            *p = 1.0f - 0.5f * (1.0f - r);
                        else
                            *p = 1.0f;
                    } else
                        *p = 0.0f;
                }
            }

            mask->unlock ();
        }
    }
}

void BrushInstrument::initBrushTexture_hairy (IArray * mask, float wet) {
    PROFILE_FUNC();
    //srand (m_LastX * m_LastY);  // FIXME Not needed anymore?

    // Hair simulation parameters.
    const int blobs_count   = MAX (1, (int)(64.0f * (1.0f - wet)));
    const int blobs_radius  = MAX (4, (int)(0.3f * mask->width () * wet));
    const int blobs_scatter = MAX (0, (int)(0.2f * mask->width () * (1.0f - wet)));

    if (mask) {
        float * p = (float*) mask->lock_pointer ();
        if (p) {
            const int w  = mask->width  ();
            const int h  = mask->height ();
            const int cx = w / 2;
            const int cy = h / 2;

            memset (p, 0, w * h * sizeof(float));

            // Scatter the blobs.
            for (int i = 0; i < blobs_count; i++) {
                const int x0 = (int)(cx + blobs_scatter * QHASMath::randGaussian ());
                const int y0 = (int)(cy + blobs_scatter * QHASMath::randGaussian ());
                const float alpha = 1.0f - MIN (1.0f, 0.5f * (abs (cx - x0) + abs (cy - y0)) / w);

                //const int r = blobs_radius *

                for (int dx = -blobs_radius; dx <= +blobs_radius; dx++) {
                    for (int dy = -blobs_radius; dy <= +blobs_radius; dy++) {
                        if (dx * dx + dy * dy > blobs_radius * blobs_radius)
                            continue;

                        const int x = x0 + dx;
                        const int y = y0 + dy;
                        if (x < 0 || x >= w ||
                            y < 0 || y >= h)
                            continue;

                        p[x + y * w] = alpha;
                    }
                }
            }

            mask->unlock ();
        }
    }
}

void BrushInstrument::copyBrushTexture (IArray * src, IArray * dst, bool alpha) {
    //srand (m_LastX * m_LastY);

    if (src && dst) {
        float * s = (float*) src->lock_pointer_const ();
        float * d = (float*) dst->lock_pointer ();
        if (s && d) {
            const int sz = src->width () * src->height ();
            if (alpha) {
                for (int i = sz; i; i--, s++, d++) {
                    *d = *s;
                }
            } else {
                for (int i = sz; i; i--, s++, d++) {
                    *d = *s > 0.5f ? 1.0f : 0.0f;
                }
            }

            src->unlock ();
            dst->unlock ();
        }
    }
}

BrushInstrument::~BrushInstrument (void) {
    ISpawnedObject::_unregister (this);
    deInit ();

    SAFE_RELEASE (m_Helper);
    SAFE_RELEASE (m_Factory);
    SAFE_RELEASE (m_Logger);
}

void BrushInstrument::validate (void) {
    if (! (m_VBOValid & VALID_VERTICES)) {
        m_VBOValid |= VALID_VERTICES;
        m_Tuft.validate ();

        const float * vtx = m_Tuft.getGeometry (m_nVertices);
        m_StrokeAngle = 0.0f;//57.295779513f * m_Tuft.getAngle ();

        // Bind and fill vertex data.
        if (m_GLinited) {
            glBindBufferARB      (GL_ARRAY_BUFFER_ARB, m_VBuf[0]);
            glBufferSubDataARB   (GL_ARRAY_BUFFER_ARB, 0, m_nVertices * 7 * sizeof(float), vtx);
            glBindBufferARB      (GL_ARRAY_BUFFER_ARB, 0);
        }
    }
}

void BrushInstrument::checkInited (void) {
    const int size = m_Type == TYPE_PENCIL ? 16 : 64;//128;

    if (!m_GLinited) {
         m_GLinited = TRUE;

        m_Mask = g_fac_SPAWN_ARRAY (
            m_Helper,
            size,
            size,
            m_Helper->proposeTextureFormat (OPENGLHELPER_TEXTURE_LUMINANCE),
            GL_FLOAT,
            IArray::FLAGS_NULL
        );
        m_MaskCtl = g_fac_SPAWN_ARRAY (
            m_Helper,
            size,
            size,
            m_Helper->proposeTextureFormat (OPENGLHELPER_TEXTURE_LUMINANCE),
            GL_FLOAT,
            IArray::FLAGS_NULL
        );
        m_Color = g_fac_SPAWN_ARRAY (
            m_Helper,
            m_Helper->sharedPatchSize (),
            m_Helper->sharedPatchSize (),
            m_Helper->proposeTextureFormat (OPENGLHELPER_TEXTURE_RGBA),
            GL_FLOAT,
            IArray::FLAGS_NULL
        );

        // Clear the color buffer.
        m_Color->zero ();

        // On-demand creation of
        // OpenGL-related objects.
        initVertexBuffer ();
        setPressure (m_Pressure, INPUT_CONTROLLER_GOD);

        // Set the default color.
        switch (m_Type) {
            case TYPE_BRUSH_ROUND:
                addMass (g_ZERO, g_ZERO, 1.0f);
                break;
            case TYPE_BRUSH_FLAT:
                addMass (g_ZERO, g_ZERO, 1.0f);
                break;
            case TYPE_PENCIL:
                addMass (g_GREY, g_GREY, 1.0f);
                break;
            case TYPE_ERASER:
                addMass (g_ZERO, g_ZERO, 1.0f);
                break;
            default:
                break;
        }
    }
}

void BrushInstrument::deInit (void) {    
    if (m_GLinited) {
        glDeleteBuffersARB (1, m_VBuf);
    }

    m_Stroke         = FALSE;
    m_PositionCached = FALSE;
    m_VBOValid       = VALID_NOTHING;

    SAFE_RELEASE (m_Mask);
    SAFE_RELEASE (m_MaskCtl);
    SAFE_RELEASE (m_Color);
    m_GLinited = FALSE;
}

void BrushInstrument::setPressure (float pressure, eInputMethod source) {
    PROFILE_FUNC();
    checkInited ();

    const float e      = 0.001f;
    const float smooth = source == INPUT_CONTROLLER_MOUSE ? 0.0f : 0.7f;

    m_Pressure = smooth * m_Pressure + (1.0f - smooth) * pressure;
    m_Pressure = MATH::CLAMP (m_Pressure, 0.0f, 1.0f);

    if (pressure < e) {
        if (m_Stroke || source == INPUT_CONTROLLER_GOD) {
            // Rebuild the texture to make the
            // pattern random for each stroke.
            switch (m_Type) {
                case TYPE_BRUSH_ROUND:
                  //initBrushTexture_round (m_MaskCtl, 0.7f, FALSE);
                    initBrushTexture_hairy (m_MaskCtl, m_WaterStorage);
                    copyBrushTexture       (m_MaskCtl, m_Mask, TRUE);
                    break;
                case TYPE_BRUSH_FLAT:
                    initBrushTexture_flat  (m_MaskCtl, 0.6f, FALSE);
                    copyBrushTexture       (m_MaskCtl, m_Mask, TRUE);
                    break;
                default:
                    initBrushTexture_round (m_MaskCtl, 0.0f, FALSE);
                    copyBrushTexture       (m_MaskCtl, m_Mask, TRUE);
                    break;
            }

            m_Stroke         = FALSE;
            m_PositionCached = FALSE;
            m_StrokeAngle    = 0.0f;
        }

        // Reset the tuft shape.
        m_Tuft.reset ();
        m_Pressure = 0.0f;
    } else {
        // Change the brush penetration level.
        m_Tuft.setGround (m_Pressure);
        m_Stroke = TRUE;
    }

    invalidate (VALID_VERTICES);
}

void BrushInstrument::setPosition (int x, int y, eInputMethod source) {
    Q_UNUSED (source);
    if (m_Stroke && m_PositionCached) {
        const float w  = 0.9f;
        const float X0 = m_LastX;
        const float Y0 = m_LastY;

        m_LastX = (int)((1.0f - w) * m_LastX + w * (float) x);
        m_LastY = (int)((1.0f - w) * m_LastY + w * (float) y);

        // FIXME Friction for eraser and flat brush is ignored.
        // FIXME Friction is totally ignored.
        if (FALSE)//m_Type == TYPE_BRUSH_ROUND || m_Type == TYPE_PENCIL)
            m_Tuft.setFriction (
                -(m_LastX - X0),
                -(m_LastY - Y0)
            );
    } else {
        m_LastX = x;
        m_LastY = y;
        m_PositionCached = TRUE;
    }

    invalidate (VALID_VERTICES);
}

void BrushInstrument::setTilt (float xtilt, float ytilt, eInputMethod source) {
    Q_UNUSED (source);

    m_Tuft.setTilt (xtilt, ytilt);
}

void BrushInstrument::applyInstrument (IChannel * channel, ICanvasDelegate * delegate) {
    checkInited ();

    BrushLock * slot = NULL;

    if (channel->uid () == CHANNELUID_Optical) {
        slot = m_LockSlots + 0;
    } else
    if (channel->uid () == CHANNELUID_Meta0) {
        slot = m_LockSlots + 1;
    } else
        return;

    if (!slot->m_Locked) {
        validate ();

        switch (m_Type) {
            case TYPE_PENCIL:
                slot->m_Sensitivity  = MATH::CLAMP_01 (0.6f - 0.4f * m_Pressure);
                slot->m_BleedingRate = 0.40f + (1.0f - 0.9f * m_Hardness) * m_Pressure;
                slot->m_SuckingRate  = 0.00f;
                break;
            case TYPE_ERASER:
                slot->m_Sensitivity  = 0.01f;
                slot->m_BleedingRate = 0.01f + 0.19f * m_Pressure;
                slot->m_SuckingRate  = 0.00f;
                break;
            default:
                slot->m_Sensitivity  = MATH::CLAMP_01 (1.0f - 0.4f * m_Pressure - 0.4f * m_WaterStorage);
                slot->m_BleedingRate = 1.00f;
                slot->m_SuckingRate  = 0.20f * m_Pressure;
                break;
        }

        slot->m_Channel        = channel;
        slot->m_ScalingFactor  = 1.0f;
        slot->m_PrimitiveMask  = m_Mask;
        slot->m_PrimitiveColor = m_Color;
        slot->m_Valid          = TRUE;
        slot->m_Locked         = TRUE;
        slot->m_Angle          = (int) m_StrokeAngle;
        slot->m_VBO            = m_VBuf[0];
        slot->m_nVBOvertices   = m_nVertices;

        delegate->touch (slot);
    }
}

IInstrument * BrushInstrument::replicate (void) {
    if (m_Factory) {
        IInstrument * replica = m_Factory->create (m_Type, m_Helper);
        if (replica) {
            if (m_GLinited) {
                const long sz = requiredBufferSize ();

                // Replicate the inner state as well.
                char * buffer = new char[sz + 64];
                if (buffer) {
                    QByteArray  arr (buffer, sz);
                    QDataStream stream (&arr, QIODevice::ReadWrite | QIODevice::Unbuffered);

                    stream.setByteOrder (QDataStream::LittleEndian);
                    stream.setVersion   (QDataStream::Qt_4_4);
                    saveData (&stream, sz);

                    stream.device ()->seek (0);
                    replica->loadData (&stream, sz);

                    SAFE_DELETE_A (buffer);
                }
            }

            return replica;
        }
    }

    return NULL;
}

void BrushInstrument::slotUnlocked (BrushLock * slot0) {
    if (slot0) {
        slot0->m_Locked = FALSE;
    }
}

void BrushInstrument::addMass (IPigmentMixture * pigment, float amount) {
    const float * a = (pigment && m_Type != TYPE_ERASER) ? pigment->absorbing ()  : g_ZERO;
    const float * s = (pigment && m_Type != TYPE_ERASER) ? pigment->scattering () : g_ZERO;

    addMass (a, s, amount);
}

void BrushInstrument::addMass (const float * a,
                               const float * s,
                               float amount) {
    PROFILE_FUNC();
    checkInited ();

    m_Changed = TRUE;

    float moist = 1.0f;
    if (m_Type == TYPE_PENCIL) {
        moist = +10.0f;
    } else
    if (m_Type == TYPE_ERASER) {
        moist = +10.0f;
    }

    // Update the channel with
    // the given values.
    if (a && s) {
        float sum_a[4] = {0.0f, 0.0f, 0.0f, 0.0f};
        float sum_s[4] = {0.0f, 0.0f, 0.0f, 0.0f};

        float * line = (float*) m_Color->lock_pointer ();
        if (line) {
            // Update each component.
            for (int i = 0; i < INSTRUMENT_COLORBUFFER_SIZE; i++,
                 line += 4 * m_Helper->sharedPatchSize ()) {

                float * pa = line;
                float * ps = line + INSTRUMENT_COLORBUFFER_SIZE * 4;
                for (int j = 0; j < INSTRUMENT_COLORBUFFER_SIZE; j++, pa += 4, ps += 4) {

                    pa[0] = (1.0f - amount) * pa[0] + amount * a[0];
                    pa[1] = (1.0f - amount) * pa[1] + amount * a[1];
                    pa[2] = (1.0f - amount) * pa[2] + amount * a[2];
                    pa[3] = (1.0f - amount) * pa[3] + amount * moist;
                    ps[0] = (1.0f - amount) * ps[0] + amount * s[0];
                    ps[1] = (1.0f - amount) * ps[1] + amount * s[1];
                    ps[2] = (1.0f - amount) * ps[2] + amount * s[2];
                    ps[3] = (1.0f - amount) * ps[3] + amount * moist;
                    if (m_Type == TYPE_PENCIL) {
                        // It seems that pencil is better emulated
                        // with low-scattering pigments.
                        ps[0] *= 0.05f;
                        ps[1] *= 0.05f;
                        ps[2] *= 0.05f;
                    }

                    sum_a[0] += pa[0];
                    sum_a[1] += pa[1];
                    sum_a[2] += pa[2];
                    sum_a[3] += pa[3];
                    sum_s[0] += ps[0];
                    sum_s[1] += ps[1];
                    sum_s[2] += ps[2];
                    sum_s[3] += ps[3];
                }
            }

            m_Color->unlock ();
        }

        // Average.
        sum_a[0] /= PRIMITIVE_COLOR_CELLS;
        sum_a[1] /= PRIMITIVE_COLOR_CELLS;
        sum_a[2] /= PRIMITIVE_COLOR_CELLS;
        sum_a[3] /= PRIMITIVE_COLOR_CELLS;
        sum_s[0] /= PRIMITIVE_COLOR_CELLS;
        sum_s[1] /= PRIMITIVE_COLOR_CELLS;
        sum_s[2] /= PRIMITIVE_COLOR_CELLS;
        sum_s[3] /= PRIMITIVE_COLOR_CELLS;

        memcpy (m_AbsorbingAvg,  sum_a, sizeof(float) * 4);
        memcpy (m_ScatteringAvg, sum_s, sizeof(float) * 4);
    }

    //m_WaterStorage = 1.0f;
    invalidate (VALID_COLORS);
}

void BrushInstrument::cleanMass (void) {
    checkInited ();
    m_Changed = TRUE;

    m_Color->zero ();

    //m_WaterStorage = 0.0f;
    memset (m_AbsorbingAvg,  0, sizeof(float) * 4);
    memset (m_ScatteringAvg, 0, sizeof(float) * 4);
    invalidate (VALID_COLORS);
}

void BrushInstrument::statePush (void) {
    sState s;
    s.x        = (float) m_LastX;
    s.y        = (float) m_LastY;
    s.pressure = m_Pressure;
    s.water    = m_WaterStorage;

    m_StatesBk.push (s);
}

void BrushInstrument::statePop (void) {
    if (!m_StatesBk.isEmpty ()) {
        sState s = m_StatesBk.pop ();

        m_LastX        = (int) s.x;
        m_LastY        = (int) s.y;
        m_Pressure     = s.pressure;
        m_WaterStorage = s.water;
    }
}

unsigned long BrushInstrument::requiredBufferSize (void) const {
    return sizeof(float) * (m_Helper->sharedPatchSize () * 4 + 1);
}

unsigned long BrushInstrument::saveData (QDataStream * stream, unsigned long capacity) {
    checkInited ();
    if (capacity < requiredBufferSize ())
        return 0L;

    // Lock the buffer.
    const char * color = (char*) m_Color->lock_pointer_const ();
    if (color) {
        const unsigned long sz = sizeof(float) * m_Helper->sharedPatchSize () * 4;
        for (unsigned long wr = 0; wr < sz; wr += stream->writeRawData (color + wr, sz - wr)) ;

        m_Color->unlock ();
    }

    stream->operator << (m_WaterStorage);

    return requiredBufferSize ();
}

unsigned long BrushInstrument::loadData (QDataStream * stream, unsigned long capacity) {
    checkInited ();
    if (capacity < requiredBufferSize ())
        return 0L;

    // Lock the buffer.
    char * color = (char*) m_Color->lock_pointer ();
    if (color) {
        const unsigned long sz = sizeof(float) * m_Helper->sharedPatchSize () * 4;
        for (unsigned long r = 0; r < sz; r += stream->readRawData (color + r, sz - r)) ;

        m_Color->unlock ();
    }

    stream->operator >> (m_WaterStorage);

    invalidate (VALID_COLORS);
    return requiredBufferSize ();
}

static const char * g_attr_Wetness  = "wetness";
static const char * g_attr_Hardness = "hardness";

/// Checks the given instrument attribute.
bool BrushInstrument::customParameterGet (const char * name, void * buffer) {
    if (m_Type == TYPE_BRUSH_FLAT ||
        m_Type == TYPE_BRUSH_ROUND) {
        if (0 == strcmp (name, g_attr_Wetness)) {
            *((float*)buffer) = m_WaterStorage;
            return TRUE;
        }
    } else
        if (m_Type == TYPE_PENCIL) {
        if (0 == strcmp (name, g_attr_Hardness)) {
            *((float*)buffer) = m_Hardness;
            return TRUE;
        }
    }

    return FALSE;
}

/// Tries to set the given instrument attribute.
bool BrushInstrument::customParameterSet (const char * name, const void * buffer) {
    if (m_Type == TYPE_BRUSH_FLAT ||
        m_Type == TYPE_BRUSH_ROUND) {
        if (0 == strcmp (name, g_attr_Wetness)) {
            m_WaterStorage = *((float*)buffer);
            return TRUE;
        }
    } else
        if (m_Type == TYPE_PENCIL) {
        if (0 == strcmp (name, g_attr_Hardness)) {
            m_Hardness = *((float*)buffer);
            return TRUE;
        }
    }

    return FALSE;
}

const char * BrushInstrument::customParameterName (int index) const {
    if (m_Type == TYPE_BRUSH_FLAT ||
        m_Type == TYPE_BRUSH_ROUND) {
        if (index == 0)
            return g_attr_Wetness;
    } else
    if (m_Type == TYPE_PENCIL) {
        if (index == 0)
            return g_attr_Hardness;
    }

    return NULL;
}

int BrushInstrument::customParametersCount (void) const {
    if (m_Type == TYPE_BRUSH_FLAT ||
        m_Type == TYPE_BRUSH_ROUND) {
        return 1;
    } else
    if (m_Type == TYPE_PENCIL) {
        return 1;
    }

    return 0;
}
