/***************************************************************************
 *  Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "rendering_chain.h"
#include <stdarg.h>

//#ifdef QT_OPENGL_LIB

static const char * g_err_OutofRange = "Maximum number of channels reached.";
static const char * g_err_WrongIndex = "Index value is out of range.";
static const char * g_err_Suicide    = "Back to Sansara...";

struct sCachedPatch {
    const sRect * prect;
    IArray      * optical_left;
    IArray      * optical_top;
    IArray      * optical_right;
    IArray      * optical_bottom;
    IArray      * optical_center;
    IArray      * meta0;

    float clip_tx0;
    float clip_tx1;
    float clip_ty0;
    float clip_ty1;

    float patch_x0;
    float patch_x1;
    float patch_y0;
    float patch_y1;
};


IRenderingChain * g_fac_SPAWN_RENDERING_CHAIN (IOpenGLHelper * helper) {
    return new RenderingChain_gl (helper);
}

void RenderingChainLayer::init (RenderingChain_gl * caller, float x, float y, float w, float h) {
    m_XOffset    = x;
    m_YOffset    = y;
    m_Width      = w;
    m_Height     = h;
    m_nChannels  = 0;
    m_Opacity    = 1.0f;
    m_Mother     = caller;
    m_HasBG      = FALSE;
    m_BGColor[0] = 0.0f;
    m_BGColor[1] = 0.0f;
    m_BGColor[2] = 0.0f;
    m_BGColor[3] = 0.0f;
}

IChannel * RenderingChainLayer::at (int index) {
    if (index >= 0 && index < m_nChannels) {
        if (m_Mother)
            m_Mother->m_LastError = NULL;
        return m_Channels[index];
    } else {
        if (m_Mother)
            m_Mother->logMessage (LOGIMPORTANCE_WARNING, g_err_WrongIndex);
        return NULL;
    }
}

void RenderingChainLayer::prepend (IChannel * channel) {
    if (m_nChannels < MAX_CHANNELS) {
        for (int i = m_nChannels; i > 0 ; i--) {
            m_Channels[i] = m_Channels[i - 1];
        }

        channel->addRef ();
        m_Channels[0] = channel;
        m_nChannels++;
        if (m_Mother)
            m_Mother->m_LastError   = NULL;
    } else {
        if (m_Mother)
            m_Mother->logMessage (LOGIMPORTANCE_ERROR, g_err_OutofRange);
    }
}

void RenderingChainLayer::append (IChannel * channel) {
    if (m_nChannels < MAX_CHANNELS) {
        channel->addRef ();
        m_Channels[m_nChannels] = channel;
        m_nChannels++;
        if (m_Mother)
            m_Mother->m_LastError   = NULL;
    } else {
        if (m_Mother)
            m_Mother->logMessage (LOGIMPORTANCE_ERROR, g_err_OutofRange);
    }
}

void RenderingChainLayer::removeAt (int index) {
    if (index >= 0 && index < m_nChannels) {
        m_Channels[index]->release ();
        for (int i = index; i < m_nChannels - 1; i++) {
            m_Channels[i] = m_Channels[i + 1];
            m_nChannels--;
        }

        if (m_Mother)
            m_Mother->m_LastError = NULL;
    } else {
        if (m_Mother)
            m_Mother->logMessage (LOGIMPORTANCE_WARNING, g_err_WrongIndex);
    }
}

void RenderingChainLayer::clear (void) {
    for (int i = 0; i < m_nChannels; i++) {
        m_Channels[i]->release ();
    }

    m_nChannels = 0;
    if (m_Mother) {
        m_Mother->m_LastError = NULL;
        m_Mother->logMessage (LOGIMPORTANCE_NEVERMIND, "Channel stack cleared.");
    }
}


RenderingChain_gl::RenderingChain_gl (IOpenGLHelper * helper) {
   m_nLayers        = 0;
   m_LastError      = NULL;
   m_Logger         = NULL;
   m_Bump           = NULL;
   m_BumpDpi        = 100;
   m_Helper         = helper;
   m_Filtering      = GL_LINEAR;
   m_glslKMlayer    = NULL;
   m_glslMVectors   = NULL;

   m_LoggingEnabled = FALSE;
   m_Options        = RENDER_PAINTS |
                      RENDER_PAPER_BUMP |
                      RENDER_PAINT_BUMP;

   // Allocate a new Id for this object.
   m_Id = ISpawnedObject::allocateId ();
   _register (this);

   // Init the light position.
   m_LightPosition[0] =   0.0f;
   m_LightPosition[1] =   0.0f;
   m_LightPosition[2] = -10.0f;

   if (helper) {
       helper->addRef ();
       reloadShaders ();
   }

   // Bump-mapping is disabled by default,
    m_PaintThickness = 0.0f;
}

void RenderingChain_gl::reloadShaders (void) {
    // Load the shaders.
    IOpenGLShader * sh = m_Helper->loadShader ("vertex_bump.glsl", "kmlayer.glsl");
    if (sh) {
        SAFE_RELEASE (m_glslKMlayer);
        m_glslKMlayer = sh;
    }

    sh = m_Helper->loadShader ("mvectors.glsl");
    if (sh) {
        SAFE_RELEASE (m_glslMVectors);
        m_glslMVectors = sh;
    }

    if (m_glslKMlayer) {
        m_glpar0_Meta0     = m_glslKMlayer->getUniformLocation ("Meta0");
        m_glpar0_PaperBump = m_glslKMlayer->getUniformLocation ("paper_bump");
        m_glpar0_BumpOff   = m_glslKMlayer->getUniformLocation ("bump_off");
        m_glpar0_BumpScale = m_glslKMlayer->getUniformLocation ("bump_scale");
        m_glpar0_I         = m_glslKMlayer->getUniformLocation ("I");
        m_glpar0_J         = m_glslKMlayer->getUniformLocation ("J");
        m_glpar0_Opacity   = m_glslKMlayer->getUniformLocation ("opacity");

        m_glpar0_KSRT_center     = m_glslKMlayer->getUniformLocation ("KSRT_center");
        m_glpar0_KSRT_left       = m_glslKMlayer->getUniformLocation ("KSRT_left");
        m_glpar0_KSRT_right      = m_glslKMlayer->getUniformLocation ("KSRT_right");
        m_glpar0_KSRT_top        = m_glslKMlayer->getUniformLocation ("KSRT_top");
        m_glpar0_KSRT_bottom     = m_glslKMlayer->getUniformLocation ("KSRT_bottom");
        m_glpar0_Paper_Embossing = m_glslKMlayer->getUniformLocation ("paper_embossing");
        m_glpar0_Paint_Embossing = m_glslKMlayer->getUniformLocation ("paint_embossing");
    }
    if (m_glslMVectors) {
        m_glpar1_Meta0     = m_glslMVectors->getUniformLocation ("Meta0");
        m_glpar1_grid      = m_glslMVectors->getUniformLocation ("grid");
        m_glpar1_color     = m_glslMVectors->getUniformLocation ("color");
    }
}

RenderingChain_gl::~RenderingChain_gl (void) {
    _unregister (this);

    clear ();
    logMessage (LOGIMPORTANCE_NEVERMIND, g_err_Suicide);

    SAFE_RELEASE (m_glslKMlayer);
    SAFE_RELEASE (m_glslMVectors);
    SAFE_RELEASE (m_Bump);
    SAFE_RELEASE (m_Helper);
    SAFE_RELEASE (m_Logger);
}

void RenderingChain_gl::setLightPosition (float x, float y, float z) {
   m_LightPosition[0] = x;
   m_LightPosition[1] = y;
   m_LightPosition[2] = z;
}

bool RenderingChain_gl::clear (void) {
    for (int i = 0; i < m_nLayers; i++) {
        m_Layers[i].clear ();
    }

    m_nLayers = 0;
    m_LastError = NULL;
    logMessage (LOGIMPORTANCE_NEVERMIND, "Rendering chain cleared.");
    return TRUE;
}

IRenderingChainLayer * RenderingChain_gl::addLayer (float x, float y, float w, float h) {
    if (m_nLayers < MAX_LAYERS) {
        m_Layers[m_nLayers].init (this, x, y, w, h);
        return m_Layers + m_nLayers++;
    } else {
        return NULL;
    }
}

/// Perform an OpenGL rendering on a given target.
bool RenderingChain_gl::render (GLuint glTarget, const sRenderOperation * operation) {
    PROFILE_FUNC();

    m_LastError = NULL;
    QMutexLocker locker (m_Helper->mutex ());

    if (!m_Helper)
        return FALSE;

    bool valid = TRUE;

    // Check some conditional stuff...
    if (operation->roi_w > operation->target_w ||
        operation->roi_h > operation->target_h ||
        operation->roi_x < 0 ||
        operation->roi_y < 0 ||
        operation->roi_w < 4 ||
        operation->roi_h < 4 ||
        operation->target_w < 4 ||
        operation->target_h < 4) {
        valid = FALSE;
    }

    if (!m_Bump ||
        !m_glslKMlayer)
        valid = FALSE;

    if (!valid) {
        // If we have some conditions unsatisfied,
        // just fill the target with grey and return.
        glDrawBuffer (glTarget);
        glClearColor (0.4f, 0.4f, 0.45f, 1.0f);
        glClear      (GL_COLOR_BUFFER_BIT);
        return TRUE;
    }

    // Setup the projection matrix.
    glMatrixMode (GL_PROJECTION);
    glPushMatrix ();
    glLoadIdentity ();
    glOrtho (-0.5f, +0.5f, -0.5f, +0.5f, +0.0f, +16.0f);
    glScalef (1.0f, 1.0f, 1.0f);

    // Setup the modelview matrix I.
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix ();
    glLoadIdentity ();
    glTranslatef (
            operation->view_translate[0],
            operation->view_translate[1],
            operation->view_translate[2] - 1.0f
    );
    glScalef (
            +operation->view_scale[0],
            -operation->view_scale[1],
            +operation->view_scale[2]
    );
    glRotatef (operation->view_rotate[0], 1.0f, 0.0f, 0.0f);
    glRotatef (operation->view_rotate[1], 0.0f, 1.0f, 0.0f);
    glRotatef (operation->view_rotate[2], 0.0f, 0.0f, 1.0f);

    float matrixbk[16];
    glGetFloatv (GL_MODELVIEW_MATRIX, matrixbk);

    glDrawBuffer (glTarget);
    glViewport   (0, 0, operation->target_w, operation->target_h);
    glEnable     (GL_SCISSOR_TEST);
    glScissor    (operation->roi_x, operation->roi_y, operation->roi_w, operation->roi_h);
    glClearColor (0.4f, 0.4f, 0.45f, 1.0f);
    glClear      (GL_COLOR_BUFFER_BIT);

    // Prepare the patches list.
    QList<sCachedPatch> patchlist;
    for (int patch_y = 0; patch_y < 32; patch_y++) {
        for (int patch_x = 0; patch_x < 32; patch_x++) {

            // For all layers...
            for (int l = 0; l < m_nLayers; l++) {
                RenderingChainLayer * layer = m_Layers + l;

              //const float layer_x0 = layer->offsetX ();
              //const float layer_y0 = layer->offsetY ();

                IChannel * ch_kmrt  = NULL;
                IChannel * ch_meta  = NULL;

                // Scan the channels list.
                for (int i = 0; i < layer->size (); i++) {
                    IChannel * channel = layer->at (i);
                    if (channel->uid () == CHANNELUID_Optical) {
                        ch_kmrt = channel;
                    } else
                    if (channel->uid () == CHANNELUID_Meta0)
                        ch_meta = channel;
                }

                // For each set of channels...
                if (ch_kmrt && ch_meta) {
                    IArrayGrid * pKMRT = ch_kmrt->lock ();
                    IArrayGrid * pMeta = ch_meta->lock ();

                    if (!pKMRT->patch (patch_x, patch_y))
                        continue;

                    const sRect * prect = pKMRT->patch (patch_x, patch_y)->rect ();

                    // Discard the off-plane patches.
                    if (prect->left >= operation->plane_width  ||
                        prect->top  >= operation->plane_height ||
                        prect->right  < 0 ||
                        prect->bottom < 0)
                        continue;

                    // Unclipped coords.
                    float patch_u_x0  = -0.5f + (float) prect->left   / operation->plane_width;
                    float patch_u_x1  = -0.5f + (float) prect->right  / operation->plane_width;
                    float patch_u_y0  = -0.5f + (float) prect->top    / operation->plane_height;
                    float patch_u_y1  = -0.5f + (float) prect->bottom / operation->plane_height;

                    // Patch coords in a world space.
                    patch_u_x0  = patch_u_x0 * matrixbk[0] + patch_u_y0 * matrixbk[4] + matrixbk[12];
                    patch_u_x1  = patch_u_x1 * matrixbk[0] + patch_u_y1 * matrixbk[4] + matrixbk[12];
                    patch_u_y0  = patch_u_x0 * matrixbk[1] + patch_u_y0 * matrixbk[5] + matrixbk[13];
                    patch_u_y1  = patch_u_x1 * matrixbk[1] + patch_u_y1 * matrixbk[5] + matrixbk[13];

                    // Patch coords in a target's space.
                    patch_u_x0  = (0.5f + patch_u_x0) * operation->target_w;
                    patch_u_x1  = (0.5f + patch_u_x1) * operation->target_w;
                    patch_u_y0  = (0.5f + patch_u_y0) * operation->target_h;
                    patch_u_y1  = (0.5f + patch_u_y1) * operation->target_h;

                    if (patch_u_y0 > patch_u_y1)
                        SWAP (patch_u_y0, patch_u_y1);

                    // Discard the patches out of ROI.
                    if (patch_u_x0 >= operation->roi_x + operation->roi_w ||
                        patch_u_y0 >= operation->roi_y + operation->roi_h ||
                        patch_u_x1 <  operation->roi_x ||
                        patch_u_y1 <  operation->roi_y)
                        continue;

                    // Clip the patch to avoid the rendering
                    // of the out-of-canvas areas.
                    const int clip_left   = MAX (prect->left,   0);
                    const int clip_top    = MAX (prect->top,    0);
                    const int clip_right  = MIN (prect->right,  operation->plane_width);
                    const int clip_bottom = MIN (prect->bottom, operation->plane_height);

                    const float TX_MARGIN = (GLEW_VERSION_1_2) ? 0.0f : 0.001f;

                    sCachedPatch p;

                    // Clipped tex-coords.
                    p.clip_tx0  = (float) (clip_left   - prect->left) / m_Helper->sharedPatchSize () + TX_MARGIN;
                    p.clip_tx1  = (float) (clip_right  - prect->left) / m_Helper->sharedPatchSize () + TX_MARGIN;
                    p.clip_ty0  = (float) (clip_top    - prect->top)  / m_Helper->sharedPatchSize () - TX_MARGIN;
                    p.clip_ty1  = (float) (clip_bottom - prect->top)  / m_Helper->sharedPatchSize () - TX_MARGIN;

                    // Clipped vertex coords.
                    p.patch_x0  = -0.5f + (float) clip_left   / operation->plane_width;
                    p.patch_x1  = -0.5f + (float) clip_right  / operation->plane_width;
                    p.patch_y0  = -0.5f + (float) clip_top    / operation->plane_height;
                    p.patch_y1  = -0.5f + (float) clip_bottom / operation->plane_height;

                    p.meta0   = pMeta->patch (patch_x, patch_y)->data ();
                    p.prect   = prect;

                    p.optical_center  = pKMRT->patch (patch_x + 0, patch_y + 0)->data ();
                    p.optical_left    = pKMRT->patch (patch_x - 1, patch_y + 0)->data ();
                    p.optical_right   = pKMRT->patch (patch_x + 1, patch_y + 0)->data ();
                    p.optical_top     = pKMRT->patch (patch_x + 0, patch_y - 1)->data ();
                    p.optical_bottom  = pKMRT->patch (patch_x + 0, patch_y + 1)->data ();

                    patchlist.append (p);
                }
            }
        }
    }

    m_glslKMlayer->useProgram ();
    m_glslKMlayer->uniform1i (m_glpar0_Meta0,        1);
    m_glslKMlayer->uniform1i (m_glpar0_KSRT_center,  3);
    m_glslKMlayer->uniform1i (m_glpar0_KSRT_left,    4);
    m_glslKMlayer->uniform1i (m_glpar0_KSRT_right,   5);
    m_glslKMlayer->uniform1i (m_glpar0_KSRT_top,     6);
    m_glslKMlayer->uniform1i (m_glpar0_KSRT_bottom,  7);

    // FIXME Layer opacity.
    m_glslKMlayer->uniform1f (m_glpar0_Opacity,   1.0f);//layer->opacity ());

    // Bump mappping.
    m_glslKMlayer->uniform1f (m_glpar0_Paint_Embossing, (m_Options & RENDER_PAINT_BUMP)
                              ? MAX (0.02f, m_PaintThickness) : 0.0f);
    m_glslKMlayer->uniform1f (m_glpar0_Paper_Embossing, (m_Options & RENDER_PAPER_BUMP)
                              ? 1.0f : 0.0f);

    // Paper color
    m_glslKMlayer->uniform4f (m_glpar0_J, 1.0f, 1.0f, 1.0f, 1.0f);

    // Light color.
    m_glslKMlayer->uniform4f (m_glpar0_I, 1.0f, 1.0f, 1.0f, 1.0f);

    // Use bump map if any.
    GLuint glBump = UINT_UNDEFINED;
    if (m_Bump) {
        glBump = m_Bump->lock_gltex ();
        glActiveTexture (GL_TEXTURE0 + 2);
        glBindTexture   (GL_TEXTURE_2D, glBump);
        glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, m_Filtering);
        glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, m_Filtering);
        glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

        float rsx = 1.0f;//(float) operation->target_w / operation->plane_width;
        float rsy = 1.0f;//(float) operation->target_h / operation->plane_height;

        // Scale the paper to image dpi.
        rsx *= (float) operation->target_dpi / m_BumpDpi;
        rsy *= (float) operation->target_dpi / m_BumpDpi;

        m_glslKMlayer->uniform2f  (m_glpar0_BumpScale,
            rsx * m_Helper->sharedPatchSize () / m_Bump->width (),
            rsy * m_Helper->sharedPatchSize () / m_Bump->height ()
        );

        m_glslKMlayer->uniform1i  (m_glpar0_PaperBump, 2);
    } else {
        m_glslKMlayer->uniform1i  (m_glpar0_PaperBump, -1);
    }

    // Process the patch list.
    // Render the scene.
    if (m_Options & RENDER_PAINTS) {
        for (int i = 0; i < patchlist.size (); i++) {
            sCachedPatch patch = patchlist[i];

            m_glslKMlayer->uniform2f  (m_glpar0_BumpOff,
                 (float) patch.clip_tx0 / m_Bump->width (), // clip_left
                 (float) patch.clip_ty0 / m_Bump->height () // clip_top
            );

            m_Helper->bindTexture_ex (patch.meta0->lock_gltex (),          1);
            m_Helper->bindTexture_ex (patch.optical_center->lock_gltex (), 3);
            m_Helper->bindTexture_ex (patch.optical_left->lock_gltex (),   4);
            m_Helper->bindTexture_ex (patch.optical_right->lock_gltex (),  5);
            m_Helper->bindTexture_ex (patch.optical_top->lock_gltex (),    6);
            m_Helper->bindTexture_ex (patch.optical_bottom->lock_gltex (), 7);

            glBegin(GL_QUADS);
               glMultiTexCoord2f (GL_TEXTURE0, patch.clip_tx0, patch.clip_ty0);
               glVertex3f (patch.patch_x0, patch.patch_y0, 1.0f);

               glMultiTexCoord2f (GL_TEXTURE0, patch.clip_tx1, patch.clip_ty0);
               glVertex3f (patch.patch_x1, patch.patch_y0, 1.0f);

               glMultiTexCoord2f (GL_TEXTURE0, patch.clip_tx1, patch.clip_ty1);
               glVertex3f (patch.patch_x1, patch.patch_y1, 1.0f);

               glMultiTexCoord2f (GL_TEXTURE0, patch.clip_tx0, patch.clip_ty1);
               glVertex3f (patch.patch_x0, patch.patch_y1, 1.0f);
            glEnd();

            patch.optical_center->unlock ();
            patch.optical_left->unlock ();
            patch.optical_right->unlock ();
            patch.optical_top->unlock ();
            patch.optical_bottom->unlock ();
            patch.meta0->unlock ();
        }
    }

    if (m_Bump)
        m_Bump->unlock ();

    // Render the motion vectors now if requested,
    if (m_glslMVectors && (m_Options & RENDER_MOTION_VECTORS)) {
        m_glslMVectors->useProgram ();
        m_glslMVectors->uniform1i (m_glpar1_Meta0, 1);
        m_glslMVectors->uniform1f (m_glpar1_grid,  0.025f);
        m_glslMVectors->uniform4f (m_glpar1_color, 0.1f, 0.8f, 0.5f, 1.0f);

        for (int i = 0; i < patchlist.size (); i++) {
            sCachedPatch patch = patchlist[i];

            m_Helper->bindTexture_ex (patch.meta0->lock_gltex (), 1);

            glBegin(GL_QUADS);
               glMultiTexCoord2f (GL_TEXTURE0, patch.clip_tx0, patch.clip_ty0);
               glVertex3f (patch.patch_x0, patch.patch_y0, 1.0f);

               glMultiTexCoord2f (GL_TEXTURE0, patch.clip_tx1, patch.clip_ty0);
               glVertex3f (patch.patch_x1, patch.patch_y0, 1.0f);

               glMultiTexCoord2f (GL_TEXTURE0, patch.clip_tx1, patch.clip_ty1);
               glVertex3f (patch.patch_x1, patch.patch_y1, 1.0f);

               glMultiTexCoord2f (GL_TEXTURE0, patch.clip_tx0, patch.clip_ty1);
               glVertex3f (patch.patch_x0, patch.patch_y1, 1.0f);
            glEnd();

            patch.meta0->unlock ();
        }
    }

#ifdef DEBUG_GLFINISH
    glFinish ();
#endif

    glDisable   (GL_SCISSOR_TEST);
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix ();

    glMatrixMode (GL_PROJECTION);
    glPopMatrix ();

    return TRUE;
}

//#endif
