/***************************************************************************
 *  Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */
#pragma once
#include "commdefs.h"
#include "commtypes.h"
#include "refcounter.h"

//#ifdef QT_OPENGL_LIB
#   include "../../../glew/include/GL/glew.h"

#include "iArray.h"
#include "iChannel.h"
#include "iOpenGLHelper.h"
#include "iRenderingChain.h"


#   define UINT_UNDEFINED   ((UINT)-1)
#   define MAX_CHANNELS     16
#   define MAX_LAYERS       64


class RenderingChain_gl;

/**
 * A single layer of a rendering ,chain.
 */
class RenderingChainLayer : public IRenderingChainLayer {
    RenderingChain_gl * m_Mother;

protected:
    /**
     * Channels stack.
     */
    IChannel * m_Channels[MAX_CHANNELS];
    int        m_nChannels;
    float      m_XOffset;
    float      m_YOffset;
    float      m_Width;
    float      m_Height;
    int        m_HasBG;
    float      m_Opacity;
    float      m_BGColor[4];
    int        m_LayerDpi;

public:

    float opacity (void) {
        return m_Opacity;
    }

    void init (RenderingChain_gl * caller, float x, float y, float w, float h);

    /***** IRENDERINGCHAINLAYER IMPLANTS *****/

    /**
     * Set the layer opacity.
     */
    virtual void setOpacity (float opacity) {
        m_Opacity = opacity;
    }

    /**
     * Set the layer dpi.
     */
    virtual void setLayerDpi (int dpi) {
        m_LayerDpi = dpi;
    }

    /**
     * Set the backgroud color for this layer.
     * @param bg Pointer to an RGBA data or NULL to unset color.
     */
    virtual void setBackgroundColor (const float * bg) {
        m_HasBG = bg != NULL;
        if (bg) {
            m_BGColor[0] = bg[0];
            m_BGColor[1] = bg[1];
            m_BGColor[2] = bg[2];
            m_BGColor[3] = bg[3];
        }
    }

    /**
     * Get the channel at given position.
     */
    virtual IChannel * at (int index);

    /**
     * X-offset of the layer.
     */
    virtual float offsetX (void) const {
        return m_XOffset;
    }

    /**
     * Y-offset of the layer.
     */
    virtual float offsetY (void) const {
        return m_YOffset;
    }

    /**
     * Layer width.
     */
    virtual float width (void) const {
        return m_Width;
    }

    /**
     * Layer height.
     */
    virtual float height (void) const {
        return m_Height;
    }

    /**
     * Check if the layer has filled
     * with background color before rendeing.
     */
    virtual const float * hasBackgroundColor (void) const {
        if (m_HasBG)
            return m_BGColor;
        else return NULL;
    }

    /**
     * Insert the channel at the
     * begining of the list.
     */
    virtual void append (IChannel * channel);

    /**
     * Insert the channel at the
     * end of the list.
     */
    virtual void prepend (IChannel * channel);

    /**
     * Remove the channel
     * at given position.
     */
    virtual void removeAt (int index);

    /**
     * Get the number of channels
     * in this layer.
     */
    virtual int size (void) const {
        return m_nChannels;
    }

    /**
     * Remove all channels from the layer.
     */
    virtual void clear (void);

    /**
     * Get the layer dpi;
     */
    virtual int dpi (void) const {
        return m_LayerDpi;
    }
};

/**
 * GPU-based rendering chain.
 */
class RenderingChain_gl : public IRenderingChain,
                          public CRefCounter {

    _VERBOSEOBJECT
    _GLEW_ENABLED

    IOpenGLHelper * m_Helper;
    IArray        * m_Bump;
    int             m_BumpDpi;
    int             m_Id;
    int             m_Options;

friend class RenderingChainLayer;
protected:
    IOpenGLShader * m_glslKMlayer;
    IOpenGLShader * m_glslMVectors;

    GLuint m_glpar0_KSRT_center;
    GLuint m_glpar0_KSRT_left;
    GLuint m_glpar0_KSRT_right;
    GLuint m_glpar0_KSRT_top;
    GLuint m_glpar0_KSRT_bottom;
    GLuint m_glpar0_Meta0;
    GLuint m_glpar0_PaperBump;
    GLuint m_glpar0_BumpOff;
    GLuint m_glpar0_BumpScale;
    GLuint m_glpar0_I;
    GLuint m_glpar0_J;
    GLuint m_glpar0_Opacity;
    GLuint m_glpar0_Paper_Embossing;
    GLuint m_glpar0_Paint_Embossing;
    GLuint m_glpar1_Meta0;
    GLuint m_glpar1_grid;
    GLuint m_glpar1_color;
    GLenum m_Filtering;

    float m_LightPosition[3];
    float m_PaintThickness;

    /**
     * Layers stack.
     */
    RenderingChainLayer m_Layers[MAX_LAYERS];
    int                 m_nLayers;

    RenderingChain_gl (IOpenGLHelper * helper);
    virtual ~RenderingChain_gl (void);

    virtual void suicide (void) {
        delete this;
    }

friend IRenderingChain * g_fac_SPAWN_RENDERING_CHAIN (IOpenGLHelper * helper);
public:

    /********** ISPAWNEDOBJECT IMPLANTS *************/

    virtual int objectId (void) {
        return m_Id;
    }

    virtual void addRef (void) {
        CRefCounter::addRef ();
    }

    virtual void release (void) {
        CRefCounter::release ();
    }

    /**
     * Ask the object to emit all
     * notification signals if any.
     */
    virtual void pokeObject (void) {
    }

    virtual const char * myClassName (void) {
        return "Server::CRenderingChain_gl";
    }

    /********** IRENDERINGCHAIN IMPLANTS *************/

    /**
     * Renders all channels onto a given raster.
     */
    virtual bool render (GLuint glTarget, const sRenderOperation * operation);

    /**
     * Create a new layer.
     * Returns an interface to a new layer that
     * you can use to pack the channels into.
     */
    virtual IRenderingChainLayer * addLayer (float x, float y, float w, float h);

    /**
     * Set the bump map that will be
     * used in final rendering iteration.
     */
    virtual void setBumpMap (IArray * map, int paperdpi) {
        SAFE_RELEASE (m_Bump);

        if (map) {
            m_Bump    = map;
            m_BumpDpi = paperdpi;

            m_Bump->addRef ();
        }
    }

    /**
     * Get the current filtering method
     * used in a rendering process.
     */
    virtual GLenum filteringMethod (void) const {
        return m_Filtering;
    }

    /**
     * Set the fltering method that
     * will be used in a rendering process.
     */
    virtual void setFilteringMethod (GLenum filtering) {
        m_Filtering = filtering;
    }

    /**
     * Remove all layers from the chain.
     */
    virtual bool clear (void);

    /**
     * Set the light position.
     * Used for bump mapping implementation.
     */
    virtual void setLightPosition (float x, float y, float z);

    /**
     * Reload all OpenGL shaders.
     */
    virtual void reloadShaders (void);

    /**
     * Set the rendered thickness of paint.
     * This controls the amount of bump mapping
     * applied to a wet chanel.
     */
    virtual void setPaintThickness (float z) {
        m_PaintThickness = z;
    }

    /**
     * Enable the given rendering options.
     */
    virtual void enableOptions (int flags) {
        m_Options |= flags;
    }

    /**
     * Disable the given rendering options.
     */
    virtual void disableOptions (int flags) {
        m_Options &= ~flags;
    }

    /**
     * Get the current rendering options.
     */
    virtual int renderingOptions (void) const {
        return m_Options;
    }

    /********** IVERBOSEOBJECT IMPLANTS ***********/

    virtual const char * lastError (void) const {
        return m_LastError;
    }

    virtual void attachLogger (ILogger * logger) {
        SAFE_RELEASE (m_Logger);

        if (logger) {
            m_Logger = logger;
            m_Logger->addRef ();
        }
    }

    virtual void enableLogging (bool enable) {
        m_LoggingEnabled = enable;
    }

    virtual bool loggingEnabled (void) {
        return m_LoggingEnabled;
    }
};

//#endif
